﻿using System;
using System.Collections.Generic;
using CeboMsrNet;
using System.IO;

/**
 * CeboMsr - C# example - Counter.
 *
 * This example shows how to use a counter. Digital I/O #1 is wired
 * to the counter, so every time this I/O is pulsed, the counter will
 * react on this event.
 *
 * Requirements: IO-1 must be connected to CNT (IO-0 on Cebo Stick).
 *
 * @author Thomas Hoppe
 */
namespace CounterExample
{
    class CounterExample
    {
        /// <summary>
        /// Configuration and running.
        /// </summary>
        /// <param name="device">Device to use for test.</param>
        private void CounterTest(Device device) 
        {
            // Create local copies, this shortens the source.
            DigitalPort dp0 = device.DigitalPorts[0];
            Counter cnt = device.Counters[0];

            // Set IO-1 as output.
            dp0.OutputEnableMask = 0x02;

            // Enable the counter.
            cnt.Enabled = true;

            // Check the counters current value.
            Console.WriteLine("Counter before starting: " + cnt.Value);

            // Pulse IO-1 3 times.
            for (int i = 0; i < 3; ++i) 
            {
                dp0.Value = 0x02;
                dp0.Value = 0x00;
            }

            // Check the counters current value.
            Console.WriteLine("Counter should be 3: " + cnt.Value);

            // Reset and ...
            cnt.Reset();

            // ... disable the counter.
            cnt.Enabled = false;

            // Check the counters current value.
            Console.WriteLine("Counter should be 0: " + cnt.Value);

            // Pulse IO-1 3 times, the counter should ignore these pulses.
            for (int i = 0; i < 3; ++i) 
            {
                dp0.Value = 0x02;
                dp0.Value = 0x00;
            }

            // Check the counters current value.
            Console.WriteLine("Counter should still be 0: " + cnt.Value);
        }

        /// <summary>
        /// The examples main method.
        /// </summary>
        private void RunExample()
        {
            try
            {
                // Search for devices ...
                IList<Device> devices = LibraryInterface.Enumerate(DeviceType.All);

                // If at least one has been found, use the first one ...
                if (devices.Count > 0)
                {
                    Device device = devices[0];

                    // Open device, nothing can be done without doing this.
                    device.Open();

                    // *** DIGITAL PORTS AND ANALOG OUTPUTS ARE MODIFIED DURING THIS ***
                    // *** SECTION, SO DO NOT CONNECT ANY PERIPHERALS WITHOUT STUDYING ***
                    // *** THE ACTIONS DONE HERE, OTHERWISE YOUR PERIPHERALS MAY BE DAMAGED !!! ***
                    CounterTest(device);

                    // Finalize device usage, this free's up the device, so it can be used
                    // again, including other applications.
                    device.Close();
                }
            }
            catch (IOException ex)
            {
                Console.WriteLine(ex.ToString());
            }
        }

        static void Main(string[] args)
        {
            new CounterExample().RunExample();
        }
    }
}
