package com.cesys.cebo.cebomsr;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

/**
 * The primary object to communicate with a CeboMsr-compatible device. Every instance holds a single
 * connection to a device. Instances are created during enumeration in {@link LibraryInterface}.enumerate().
 *
 * @author Thomas Hoppe
 * @throws IllegalArgumentException If the underlying API detects an illegal argument.
 * @throws IndexOutOfBoundsException If the underlying API was forced to access an invalid array element.
 * @throws IllegalStateException If a method is called that should not be called in the current state.
 * @throws IOException Unrecoverable error in the lower API levels, in most cases an unpredictable communication error with the hardware.
 * @throws RuntimeException If the API reports an unknown error, this should never be occur except for the case of a bug inside the API.
 */
public final class Device {
	private final String _ident, _firmwareVersion;
	private final DeviceType _msrType;
	private int _handle = -1;
	private int _maxInputFrames = -1;

	private final List<AnalogInput> _singleEnded = new ArrayList<AnalogInput>();
	private final List<AnalogInput> _differential = new ArrayList<AnalogInput>();
	private final List<AnalogOutput> _analogOutputs = new ArrayList<AnalogOutput>();
	private final List<DigitalPort> _digitalPorts = new ArrayList<DigitalPort>();
	private final List<Counter> _counters = new ArrayList<Counter>();
	private final List<Trigger> _triggers = new ArrayList<Trigger>();
	private final List<Led> _leds = new ArrayList<Led>();
	private final List<CurrentSource> _currentSources = new ArrayList<CurrentSource>();

	private int getSingleEndedCount() {
		return _singleEnded.size();
	}

	private int getDifferentialCount() {
		return _differential.size();
	}

	private int getAnalogOutputCount() {
		return _analogOutputs.size();
	}

	private int getDigitalPortCount() {
		return _digitalPorts.size();
	}

	private int getCounterCount() {
		return _counters.size();
	}

	private int getTriggerCount() {
		return _triggers.size();
	}

	private int getInputFrameFloatCount() {
		return getSingleEndedCount() + getDifferentialCount();
	}

	private int getInputFrameIntCount() {
		return getDigitalPortCount() + getCounterCount() + getTriggerCount();
	}

	private int getOutputFrameFloatCount() {
		return getAnalogOutputCount();
	}

	private int getOutputFrameIntCount() {
		return getDigitalPortCount();
	}

	private int getMaxInputFrames() {
		return _maxInputFrames;
	}

	/**
	 * As Java has no uint32, the parameter must be a long, but it must result
	 * in an int, that holds the binary version of the uint32. Additional,
	 * an range check is done.
	 * @param value
	 * @return
	 */
	private static int toUint32(long value) {
		final long maxValue = 4294967295L;
		final long minValue = 0;
		if (value < minValue || value > maxValue)
			throw new IllegalArgumentException("Value outside range of 32 bit unsigned int.");
		return (int)value;
	}

	/**
	 * Convert int to 32 bit unsigned representation in a long.
	 * @param value
	 * @return
	 */
	private static long fromUInt32(int value) {
		if (value >= 0)
			return (long)value;
		return 0x100000000L + (long)value;
	}

	/**
	 * Extract frames from both given buffers.
	 * @param vbuffer
	 * @param buffer
	 * @param count
	 * @return
	 */
	private List<InputFrame> convertArraysToFrames(float[] vbuffer, int[] buffer, int count) {
		List<InputFrame> r = new ArrayList<InputFrame>();
		int voffset = 0, offset = 0;
		for (int i = 0;	i < count; ++i,
				voffset += getInputFrameFloatCount(),
				offset += getInputFrameIntCount()) {
			r.add(
					new InputFrame(
							vbuffer, buffer,
							voffset, getSingleEndedCount(),
							voffset + getSingleEndedCount(), getDifferentialCount(),
							offset, getDigitalPortCount(),
							offset + getDigitalPortCount(), getCounterCount(),
							offset + getDigitalPortCount() + getCounterCount(), getTriggerCount()));
		}
		return r;
	}

	/**
	 * Read device configuration parameter.
	 * @param param
	 * @return
	 * @throws IOException
	 */
	private int getDeviceParameter(int param) throws IOException {
		int[] tmp = new int[1];
		SystemConnector.checkError(SystemConnector.CeboMsrGetDeviceParameter(_handle, param, tmp));
		return tmp[0];
	}

	int getHandle() {
		return _handle;
	}


	/**
	 * @param ident
	 * @param enumId
	 */
	Device(String ident, DeviceType msrType, String firmwareVersion) {
		_ident = ident;
		_msrType = msrType;
		_firmwareVersion = firmwareVersion;
	}

	/**
	 * Open device.
	 */
	public void open() throws IOException {
		//_logger.debug("Open device (enumId: {})", _enumId);

		int[] handleb = new int[1];
		byte[] id = _ident.getBytes();
		byte[] idnt = new byte[id.length + 1];
		System.arraycopy(id, 0, idnt, 0, id.length);
		SystemConnector.checkError(SystemConnector.CeboMsrOpen(idnt, handleb));
		_handle = handleb[0];

		_singleEnded.clear();
		for (int i = 0; i < getDeviceParameter(SystemConnector.CeboParameterSingleEndedCount); ++i)
			_singleEnded.add(new AnalogInput(this, SystemConnector.CeboMsrIoIdSingleEnded0 + i, AnalogInput.Style.SingleEnded, i));

		_differential.clear();
		for (int i = 0; i < getDeviceParameter(SystemConnector.CeboParameterDifferentialCount); ++i)
			_differential.add(new AnalogInput(this, SystemConnector.CeboMsrIoIdDifferential0 + i, AnalogInput.Style.Differential, i));

		_analogOutputs.clear();
		for (int i = 0; i < getDeviceParameter(SystemConnector.CeboParameterAnalogOutputCount); ++i)
			_analogOutputs.add(new AnalogOutput(this, SystemConnector.CeboMsrIoIdAnalogOutput0 + i, i));

		_digitalPorts.clear();
		for (int i = 0; i < getDeviceParameter(SystemConnector.CeboParameterDigitalPortCount); ++i)
			_digitalPorts.add(new DigitalPort(this, SystemConnector.CeboMsrIoIdDigitalPort0 + i, i));

		_counters.clear();
		for (int i = 0; i < getDeviceParameter(SystemConnector.CeboParameterCounterCount); ++i)
			_counters.add(new Counter(this, SystemConnector.CeboMsrIoIdCounter0 + i, i));

		_triggers.clear();
		for (int i = 0; i < getDeviceParameter(SystemConnector.CeboParameterTriggerCount); ++i)
			_triggers.add(new Trigger(this, SystemConnector.CeboMsrIoIdTrigger0 + i, i));

		_currentSources.clear();
		for (int i = 0; i < getDeviceParameter(SystemConnector.CeboParameterCurrentSourceCount); ++i)
			_currentSources.add(new CurrentSource(this, SystemConnector.CeboMsrCurrentSource0 + i, i));

		_leds.clear();
		for (int i = 0; i < getDeviceParameter(SystemConnector.CeboParameterLedCount); ++i)
			_leds.add(new Led(this, SystemConnector.CeboMsrLedIndicator + i, 0));
		
		_maxInputFrames = getDeviceParameter(SystemConnector.CeboParameterMaxInputFrames);
	}

	/**
	 * Close device.
	 */
	public void close() {
		SystemConnector.CeboMsrClose(_handle);
		_handle = -1;
	}

	/**
	 * Restore startup defaults in device.
	 */
	public void resetDevice() throws IOException {
		SystemConnector.checkError(SystemConnector.CeboMsrResetDevice(_handle));
	}

	/**
	 * Reset specified elements in device.
	 * @param mask Bit vector constructed of {@link Device}.FlagReset*.
	 */
	public void resetPeripherals(int mask) throws IOException {
		SystemConnector.checkError(SystemConnector.CeboMsrResetPeripherals(_handle, mask));
	}

	/**
	 * Return unique id of device.
	 */
	public String getIdentifier() {
		return _ident;
	}

	/**
	 * Get actual type of device.
	 */
	public DeviceType getDeviceType() {
		return _msrType;
	}

	/**
	 * Read string which contains the firmware version.
	 */
	public String getFirmwareVersion() throws IOException {
		return _firmwareVersion;
	}

	/**
	 * Return serial number of device.
	 */
	public String getSerialNumber() throws IOException {
		byte[] tmp = new byte[SystemConnector.MaxIdentLength];
		SystemConnector.checkError(SystemConnector.CeboMsrGetSerialNumber(_handle, tmp));
		return Utils.convertByteArrayToString(tmp);
	}

	/**
	 * Read device temperature in degree celsius.
	 */
	public float getTemperature() throws IOException {
		float[] fb = new float[1];
		SystemConnector.checkError(SystemConnector.CeboMsrGetTemperature(_handle, fb));
		return fb[0];
	}

	/**
	 * Calculate maximum count of frames that can be buffered on the device when using
	 * multiple frame mode. Depends on the actively selected inputs.
	 * @return Frame count.
	 */
	public int calculateMaxBufferedInputFrames() throws IOException {
		int[] r = new int[1];
		SystemConnector.checkError(SystemConnector.CeboMsrCalculateMaxBufferedInputFrames(_handle, r));
		return r[0];
	}

	/**
	 * Define the inputs that were included in the next frame request
	 * (readFrame(), readBlocking() and readNonBlocking()).
	 * @param inputs Collection of inputs.
	 */
	public void setupInputFrame(Input[] inputs) throws IOException {
		int[] transformed = new int[inputs.length];

		for (int i = 0; i < inputs.length; ++i) {
			if (inputs[i] instanceof AnalogInput)
				transformed[i] = ((AnalogInput)inputs[i]).getInterfaceId();
			else if (inputs[i] instanceof DigitalPort)
				transformed[i] = ((DigitalPort)inputs[i]).getInterfaceId();
			else if (inputs[i] instanceof Trigger)
				transformed[i] = ((Trigger)inputs[i]).getInterfaceId();
			else if (inputs[i] instanceof Counter)
				transformed[i] = ((Counter)inputs[i]).getInterfaceId();
			else {
				throw new IllegalArgumentException("Unknown input.");
			}
		}
		SystemConnector.checkError(SystemConnector.CeboMsrSetupInputFrame(_handle, transformed, transformed.length));
	}

	/**
	 * Initialize a buffered data acquisition. Frames are buffered in the device memory,
	 * so frame loss is not possible.
	 * @param frameRate Frequency in Hz to capture frame.
	 * @param frameCount Maximum count of frames to capture. Maximum possible frames can
	 *                   calculated using calculateMaxBufferedInputFrames().
	 * @param externalStarted If true, sampling starts if an external trigger is detected,
	 *                        immediately otherwise.
	 * @return Corrected frame rate.
	 */
	public float startBufferedDataAcquisition(float frameRate, int frameCount, boolean externalStarted) throws IOException {
		float[] r = new float[1];
		if (frameRate <= 0)
			throw new IllegalArgumentException("frameRate must be > 0");
		if (frameCount <= 0)
			throw new IllegalArgumentException("frameCount must be > 0");
		SystemConnector.checkError(SystemConnector.CeboMsrStartDataAcquisition(_handle, externalStarted ? 1 : 0, frameRate, frameCount, r));
		return r[0];
	}

	/**
	 * Initialize a buffered data acquisition. Frames are buffered in the device memory,
	 * so frame loss is not possible. For every detected trigger, one frame is captured.
	 * Too high frequency on the external trigger will create frame loss.
	 * @param frameCount Maximum count of frames to capture. Maximum possible frames can calculated
	 *                   using calculateMaxBufferedInputFrames().
	 */
	public void startBufferedExternalTimedDataAcquisition(int frameCount) throws IOException {
		float[] r = new float[1];
		if (frameCount <= 0)
			throw new IllegalArgumentException("frameCount must be > 0");
		SystemConnector.checkError(SystemConnector.CeboMsrStartDataAcquisition(_handle, 1, -1, frameCount, r));
	}

	/**
	 * Initialize a continuous data acquisition. Data must be read as fast as possible using
	 * readBlocking()/readNonBlocking(), otherwise  an device buffer overflow may occur.
	 * @param frameRate Frequency in Hz to capture frame.
	 * @param externalStarted If true, sampling starts if an external trigger is detected,
	 *                        immediately otherwise.
	 * @return Corrected frame rate.
	 */
	public float startContinuousDataAcquisition(float frameRate, boolean externalStarted) throws IOException {
		float[] r = new float[1];
		if (frameRate <= 0)
			throw new IllegalArgumentException("frameRate must be > 0");
		SystemConnector.checkError(SystemConnector.CeboMsrStartDataAcquisition(_handle, externalStarted ? 1 : 0, frameRate, -1, r));
		return r[0];
	}

	/**
	 * Initialize a continuous data acquisition. Data must be read as fast as possible
	 * using readBlocking()/readNonBlocking(), otherwise an device buffer overflow may occur.
	 * For every detected trigger, one frame is captured. Too high frequency on the external
	 * trigger will create frame loss.
	 */
	public void startContinuousExternalTimedDataAcquisition() throws IOException {
		float[] r = new float[1];
		SystemConnector.checkError(SystemConnector.CeboMsrStartDataAcquisition(_handle, 1, -1, -1, r));
	}

	/**
	 * Stop any form of hardware managed data acquisition.
	 */
	public void stopDataAcquisition() throws IOException {
		SystemConnector.checkError(SystemConnector.CeboMsrStopDataAcquisition(_handle));
	}

	/**
	 * Read exact frameCount frames, return after completion, blocks until all data is available.
	 * @param frameCount Count of frames to read.
	 * @return Collection of read frames.
	 */
	public List<InputFrame> readBlocking(int frameCount) throws IOException {
		float[] vbuffer = new float[getInputFrameFloatCount() * frameCount];
		int[] buffer = new int[getInputFrameIntCount() * frameCount];
		SystemConnector.checkError(SystemConnector.CeboMsrReadBlocking(_handle, vbuffer, buffer, frameCount));
		return convertArraysToFrames(vbuffer, buffer, frameCount);
	}

	/**
	 * Read at much possible available frames, return immediately.
	 * @return Collection of frames available at the moment of reading, 0 <= count <= max possible.
	 */
	public List<InputFrame> readNonBlocking() throws IOException {
		float[] vbuffer = new float[getInputFrameFloatCount() * getMaxInputFrames()];
		int[] buffer = new int[getInputFrameIntCount() * getMaxInputFrames()];
		int[] countb = new int[1];
		SystemConnector.checkError(SystemConnector.CeboMsrReadNonBlocking(_handle, vbuffer, buffer, countb));
		return convertArraysToFrames(vbuffer, buffer, countb[0]);
	}

	/**
	 * Read multiple inputs using the previously defined list (setupInputFrame()).
	 * @return Frame that contains the values of all specified inputs.
	 */
	public InputFrame readFrame() throws IOException {
		float[] vbuffer = new float[getInputFrameFloatCount()];
		int[] buffer = new int[getInputFrameIntCount()];
		SystemConnector.checkError(SystemConnector.CeboMsrReadFrame(_handle, vbuffer, buffer));
		return
				new InputFrame(
						vbuffer, buffer,
						0, getSingleEndedCount(),
						getSingleEndedCount(), getDifferentialCount(),
						0, getDigitalPortCount(),
						getDigitalPortCount(), getCounterCount(),
						getDigitalPortCount() + getCounterCount(), getTriggerCount());
	}

	/**
	 * Define the outputs that should be set using the next frame based out transfer.
	 * @param outputs Collection of outputs.
	 */
	public void setupOutputFrame(Output[] outputs) throws IOException {
		int[] transformed = new int[outputs.length];
		for (int i = 0; i < outputs.length; ++i) {
			if (outputs[i] instanceof AnalogOutput)
				transformed[i] = ((AnalogOutput)outputs[i]).getInterfaceId();
			else if (outputs[i] instanceof DigitalPort)
				transformed[i] = ((DigitalPort)outputs[i]).getInterfaceId();
			else
				throw new IllegalArgumentException("Unknown output.");
		}
		SystemConnector.checkError(SystemConnector.CeboMsrSetupOutputFrame(_handle, transformed, transformed.length));
	}

	/**
	 * Write multiple outputs at once. All outputs specified using setupOutputFrame() are modified using
	 * the values in the frame.
	 * @param frame Instance that holds all values to set.
	 */
	public void writeFrame(OutputFrame frame) throws IOException {
		SystemConnector.checkError(SystemConnector.CeboMsrWriteFrame(_handle, frame.getAnalogValues(), frame.getDigitalValues()));
	}

	/**
	 * Create OutputFrame instance specific to this device.
	 * @return New instance of OutputFrame.
	 */
	public OutputFrame createOutputFrame() {
		return new OutputFrame(getOutputFrameFloatCount(), getOutputFrameIntCount());
	}

	/**
	 * Return list of single ended analog inputs.
	 */
	public List<AnalogInput> getSingleEndedInputs() {
		return Collections.unmodifiableList(_singleEnded);
	}

	/**
	 * Return list of differential analog inputs.
	 */
	public List<AnalogInput> getDifferentialInputs() {
		return Collections.unmodifiableList(_differential);
	}

	/**
	 * Return list of analog outputs.
	 */
	public List<AnalogOutput> getAnalogOutputs() {
		return Collections.unmodifiableList(_analogOutputs);
	}

	/**
	 * Return list of digital ports.
	 */
	public List<DigitalPort> getDigitalPorts() {
		return Collections.unmodifiableList(_digitalPorts);
	}

	/**
	 * Return list of counters.
	 */
	public List<Counter> getCounters() {
		return Collections.unmodifiableList(_counters);
	}

	/**
	 * Return list of triggers.
	 */
	public List<Trigger> getTriggers() {
		return Collections.unmodifiableList(_triggers);
	}

	/**
	 * Return list of current sources.
	 */
	public List<CurrentSource> getCurrentSources() {
		return Collections.unmodifiableList(_currentSources);
	}

	/**
	 * Return list of LED's.
	 */
	public List<Led> getLeds() {
		return Collections.unmodifiableList(_leds);
	}

	/**
	 * Check if the CeboMsr API is available.
	 * (If compatible system library can be found.)
	 */
	public static boolean isSystemAvailable() {
		if (SystemConnector.isAvailable())
			return true;
		return false;
	}

	/**
	 * *** Internal use only ***
	 */
	public void readBlock(long address, byte[] buffer) throws IOException {
		readBlock(address, buffer, buffer.length);
	}

	/**
	 * *** Internal use only ***
	 */
	public void readBlock(long address, byte[] buffer, int size) throws IOException {
		SystemConnector.checkError(SystemConnector.__CeboMsrReadBlock(_handle, toUint32(address), buffer, size));
	}

	/**
	 * *** Internal use only ***
	 */
	public long readRegister(long address) throws IOException {
		int[] r = new int[1];
		SystemConnector.checkError(SystemConnector.__CeboMsrReadRegister(_handle, toUint32(address), r));
		return fromUInt32(r[0]);
	}

	/**
	 * *** Internal use only ***
	 */
	public void writeBlock(long address, byte[] buffer) throws IOException {
		writeBlock(address, buffer, buffer.length);
	}

	/**
	 * *** Internal use only ***
	 */
	public void writeBlock(long address, byte[] buffer, int size) throws IOException {
		SystemConnector.checkError(SystemConnector.__CeboMsrWriteBlock(_handle, toUint32(address), buffer, size));
	}

	/**
	 * *** Internal use only ***
	 */
	public void writeRegister(long address, long value) throws IOException {
		SystemConnector.checkError(SystemConnector.__CeboMsrWriteRegister(_handle, toUint32(address), toUint32(value)));
	}

	public String toString() {
		return "CeboMsr-Device: " + _ident;
	}

	/**
	 * Valid flags to use with {@link Device}.resetPeripherals().
	 */
	public final static int FlagResetInputFifo = SystemConnector.CeboMsrPeripheralResetMaskInputFifo;
}
