package com.cesys.cebo.cebomsr.examples;

import java.io.IOException;
import java.util.List;

import com.cesys.cebo.cebomsr.Device;
import com.cesys.cebo.cebomsr.DeviceType;
import com.cesys.cebo.cebomsr.DigitalPort;
import com.cesys.cebo.cebomsr.Counter;
import com.cesys.cebo.cebomsr.LibraryInterface;

/**
 * CeboMsr - Java example - Counter.
 *
 * This example shows how to use a counter. Digital I/O #1 is wired
 * to the counter, so every time this I/O is pulsed, the counter will
 * react on this event.
 *
 * Requirements: IO-1 must be connected to CNT (IO-0 on Cebo Stick).
 *
 * @author Thomas Hoppe
 */
public class CounterExample {

	/**
	 * Configuration and running.
	 * @param device Device to use for test.
	 * @throws IOException
	 */
	private void counterTest(Device device) throws IOException {
		// Create local copies, this shortens the source.
		DigitalPort dp0 = device.getDigitalPorts().get(0);
		Counter cnt = device.getCounters().get(0);

		// Set IO-1 as output.
		dp0.setOutputEnableMask(0x02);

		// Enable the counter.
		cnt.setEnabled(true);

		// Check the counters current value.
		System.out.println("Counter before starting: " + cnt.read());

		// Pulse IO-1 3 times.
		for (int i = 0; i < 3; ++i) {
			dp0.write(0x02);
			dp0.write(0x00);
		}

		// Check the counters current value.
		System.out.println("Counter should be 3: " + cnt.read());

		// Reset and ...
		cnt.reset();

		// ... disable the counter.
		cnt.setEnabled(false);

		// Check the counters current value.
		System.out.println("Counter should be 0: " + cnt.read());

		// Pulse IO-1 3 times, the counter should ignore these pulses.
		for (int i = 0; i < 3; ++i) {
			dp0.write(0x02);
			dp0.write(0x00);
		}

		// Check the counters current value.
		System.out.println("Counter should still be 0: " + cnt.read());
	}

	/**
	 * The examples main method.
	 */
	private void runExample() {
		try {
			// Search for devices ...
			List<Device> devices = LibraryInterface.enumerate(DeviceType.All);

			// If at least one has been found, use the first one ...
			if (!devices.isEmpty()) {
				Device device = devices.get(0);

				// Open device, nothing can be done without doing this.
				device.open();

				// *** DIGITAL PORTS AND ANALOG OUTPUTS ARE MODIFIED DURING THIS ***
				// *** SECTION, SO DO NOT CONNECT ANY PERIPHERALS WITHOUT STUDYING ***
				// *** THE ACTIONS DONE HERE, OTHERWISE YOUR PERIPHERALS MAY BE DAMAGED !!! ***
				counterTest(device);

				// Finalize device usage, this free's up the device, so it can be used
				// again, including other applications.
				device.close();
			}
		}
		catch (IOException ex) {
			ex.printStackTrace();
		}
	}

	public static void main(String[] args) {
		new CounterExample().runExample();
	}
}
