package com.cesys.cebo.cebomsr.examples;

import java.io.IOException;
import java.util.List;

import com.cesys.cebo.cebomsr.AnalogInput;
import com.cesys.cebo.cebomsr.CurrentSource;
import com.cesys.cebo.cebomsr.Device;
import com.cesys.cebo.cebomsr.DeviceType;
import com.cesys.cebo.cebomsr.DigitalPort;
import com.cesys.cebo.cebomsr.LibraryInterface;
import com.cesys.cebo.cebomsr.Range;

/**
 * CeboMsr - Java example - Info.
 *
 * Print out various information about the device.
 *
 * @author Thomas Hoppe
 */
public class InfoExample {

	/**
	 * Print various information to the console.
	 * @param device Device to use for test.
	 * @throws IOException
	 */
	private void printInformation(Device device) throws IOException {
		// Put out some device specific information.
		System.out.println("Device type: " + device.getDeviceType().getName());
		System.out.println("USB base ver: " + LibraryInterface.getUsbBaseVersion());
		System.out.println("API vers: " + LibraryInterface.getApiVersion());
		System.out.println("Firmware ver: " + device.getFirmwareVersion());
		System.out.println("Identifier: " + device.getIdentifier());
		System.out.println("Serial number: " + device.getSerialNumber());
		System.out.println("Temperature: " + device.getTemperature());

		// Get the real current of the reference current sources.
		for (CurrentSource source : device.getCurrentSources()) {
			System.out.println("Reference current of "
				+ source.getName() + ": "
				+ source.getReferenceCurrent() + " uA");
		}

		// Retrieve information about single ended input #0.
		// This works on all analog inputs and outputs.
		AnalogInput se0 = device.getSingleEndedInputs().get(0);
		System.out.println("Info for single ended input " + se0.getName() + ":");
		System.out.println("  Min. ICD: " + se0.getMinInterChannelDelay() + " us");
		System.out.println("  Current range: "
			+ se0.getRange().getMinValue() + " V to "
			+ se0.getRange().getMaxValue() + " V");
			System.out.println("Supported ranges:");
		for (int i = 0; i < se0.getSupportedRanges().size(); ++i) {
			Range range = se0.getSupportedRanges().get(i);
			System.out.println("    Range #" + i + " range: "
				+ range.getMinValue() + " V to "
				+ range.getMaxValue() + " V, "
				+ "Def. ICD: "
				+ se0.getDefaultInterChannelDelay(range) + " us");
		}

		// Get info for digital port #0.
		DigitalPort dp0 = device.getDigitalPorts().get(0);
		System.out.println("Info for digital port " + dp0.getName() + ":");
		System.out.println("  Count of I/O's: " + dp0.getIoCount());
		for (int i = 0; i < dp0.getIoCount(); ++i) {
			System.out.println("    I/O #" + i + ":" + dp0.getIoName(i));
		}
	}

	/**
	 * The examples main method.
	 */
	private void runExample() {
		try {
			// Search for devices ...
			List<Device> devices = LibraryInterface.enumerate(DeviceType.All);

			// If at least one has been found, use the first one ...
			if (!devices.isEmpty()) {
				Device device = devices.get(0);

				// Open device, nothing can be done without doing this.
				device.open();

				// Put out info.
				printInformation(device);

				// Finalize device usage, this free's up the device, so it can be used
				// again, including other applications.
				device.close();
			}
		}
		// These exceptions are currently used to report different problems ...
		catch (IllegalArgumentException ex) {
			ex.printStackTrace();
		}
		catch (IndexOutOfBoundsException ex) {
			ex.printStackTrace();
		}
		catch (IllegalStateException ex) {
			ex.printStackTrace();
		}
		catch (IOException ex) {
			ex.printStackTrace();
		}
	}

	public static void main(String[] args) {
		new InfoExample().runExample();
	}
}
